'''
Jan Habscheid
Jan.Habscheid@rwth-aachen.de

This script implements the fenics solver for a ternary electrolyte (N=3, A,C,S)
'''

import numpy as np
from mpi4py import MPI
from dolfinx import mesh, fem, log
from dolfinx.fem.petsc import NonlinearProblem, LinearProblem
from dolfinx.nls.petsc import NewtonSolver
from ufl import TestFunctions, split, dot, grad, dx, inner, ln, Mesh, div, exp
from basix.ufl import element, mixed_element
import matplotlib.pyplot as plt


def solve_System(n_left, n_right, np_left, nm_right, D0, km, kp, kn, A, N0, NL, dt, t_end, number_cells, relax_param:float=None, x0:float=0, x1:float=20, rtol:float=1e-8, max_iter:float=500):
    L = x1 - x0
    # Define boundaries for the boundary conditions
    def Left(x):
        return np.isclose(x[0], x0)

    def Right(x):
        return np.isclose(x[0], x1)
    
    # Create mesh
    msh = mesh.create_interval(MPI.COMM_WORLD, number_cells, [x0, x1], dtype=np.float64)

    # Define Finite Elements
    CG1_elem = element('Lagrange', msh.basix_cell(), 1)

    # Define Mixed Function Space
    W_elem = mixed_element([CG1_elem, CG1_elem, CG1_elem])
    W = fem.functionspace(msh, W_elem)

    # Define Trial- and Testfunctions
    u = fem.Function(W)
    u_last = fem.Function(W)
    u_n, u_np, u_nm = split(u)
    u_n_last, u_np_last, u_nm_last = split(u_last)
    (v_n, v_np, v_nm) = TestFunctions(W)

    # Collapse function space for bcs
    W0, _ = W.sub(0).collapse()
    W1, _ = W.sub(1).collapse()
    W2, _ = W.sub(2).collapse()

    # Define boundary conditions values
    def u_n_left(x):
        return np.full_like(x[0], n_left)
    def u_n_right_(x):
        return np.full_like(x[0], n_right)
    def u_np_left_(x):
        return np.full_like(x[0], np_left)
    def u_nm_right_(x):
        return np.full_like(x[0], nm_right)

    # Interpolate bcs functions
    u_n_left_bcs = fem.Function(W0)
    u_n_left_bcs.interpolate(u_n_left)
    u_n_right_bcs = fem.Function(W0)
    u_n_right_bcs.interpolate(u_n_right_)
    u_np_left_bcs = fem.Function(W1)
    u_np_left_bcs.interpolate(u_np_left_)
    u_nm_right_bcs = fem.Function(W2)
    u_nm_right_bcs.interpolate(u_nm_right_)

    # Identify dofs for boundary conditions
    # Define boundary conditions
    facet_left_dofs = fem.locate_dofs_geometrical((W.sub(0), W.sub(0).collapse()[0]), Left)
    facet_right_dofs = fem.locate_dofs_geometrical((W.sub(0), W.sub(0).collapse()[0]), Right)
    bc_left_n = fem.dirichletbc(u_n_left_bcs, facet_left_dofs, W.sub(2))
    bc_right_n = fem.dirichletbc(u_n_right_bcs, facet_right_dofs, W.sub(2))

    facet_left_dofs = fem.locate_dofs_geometrical((W.sub(1), W.sub(1).collapse()[0]), Left)
    bc_left_np = fem.dirichletbc(u_np_left_bcs, facet_left_dofs, W.sub(1))

    facet_right_dofs = fem.locate_dofs_geometrical((W.sub(2), W.sub(2).collapse()[0]), Right)
    bc_right_nm = fem.dirichletbc(u_nm_right_bcs, facet_right_dofs, W.sub(2))

    # Combine boundary conditions into list
    bcs = [bc_left_n, bc_right_n, bc_left_np, bc_right_nm]
        
    

    # Variational Form
    # A = (
    #     (u_n - u_n_last) / dt * v_n * dx + D0 * inner(grad(u_n), grad(v_n)) * dx
    # ) + (
    #     (u_np - u_np_last) / dt * v_np * dx + inner(grad(u_np)[0], v_np) * dx
    # ) + (
    #     (u_nm - u_nm_last) / dt * v_nm * dx - inner(grad(u_nm)[0], v_nm) * dx
    # )

    # exponential form 
    eps = 1e-3
    A = (
        (u_n - u_n_last) / dt * v_n * dx + D0 * inner(grad(u_n), grad(v_n)) * dx
    ) + (
        (u_np - u_np_last) / dt * v_np * dx + inner(grad(u_np * exp(-A*u_np))[0], v_np) * dx
        # + epsilon u_xx for artifical diffusion
        + eps * inner(grad(u_np), grad(v_np)) * dx
    ) + (
        (u_nm - u_nm_last) / dt * v_nm * dx + inner(grad(-u_nm * exp(-A*u_nm))[0], v_nm) * dx
        # + epsilon u_xx for artifical diffusion
        + eps * inner(grad(u_nm), grad(v_nm)) * dx
    )

    # Source Terms
    A += (
        (1 + km) * u_n * v_n * dx
    ) + (
        -u_n * v_np * dx
    ) + (
        -km * u_n * v_nm * dx
    )

    # Reaction Terms
    A += (
        - kp * u_np * v_n * dx - kn * u_nm * v_n * dx
    ) + (
        + kp * u_np * v_np * dx
    ) + (
        + kn * u_nm * v_nm * dx
    )

    # # Reaction Terms with third closure
    # ! Careful, this is work in progress and does not work yet
    # B = 75
    # A += (
    #     - kp * exp(B * u_np) * u_np * v_n * dx - kn * exp(B * u_nm) * u_nm * v_n * dx
    # ) + (
    #     + kp * exp(B * u_np) * u_np * v_np * dx
    # ) + (
    #     + kn * exp(B * u_nm) * u_nm *v_nm * dx
    # )
    
    F = A

    # Initialize initial guess for u
    n_init = fem.Function(W0)
    np_init = fem.Function(W1)
    nm_init = fem.Function(W2)
    n_init.interpolate(lambda x: N0 + (NL - N0) * x[0] / L)
    np_init.interpolate(lambda x: np.zeros_like(x[0]))
    nm_init.interpolate(lambda x: np.zeros_like(x[0]))

    with u.vector.localForm() as u_loc:
        u_loc.set(0)
    u.sub(0).interpolate(n_init)
    u.sub(1).interpolate(np_init)
    u.sub(2).interpolate(nm_init)
    with u_last.vector.localForm() as u_loc:
        u_loc.set(0)
    u_last.sub(0).interpolate(n_init)
    u_last.sub(1).interpolate(np_init)
    u_last.sub(2).interpolate(nm_init)

    # Define Nonlinear Problem
    problem = NonlinearProblem(F, u, bcs=bcs)
    # problem = LinearProblem(F, 0, bcs=bcs, u=u, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    # problem = LinearProblem(F, 0, bcs=bcs, u=u)

    # Define Newton Solver
    solver = NewtonSolver(MPI.COMM_WORLD, problem)
    solver.report = False#True

    log.set_log_level(log.LogLevel.INFO)

    cur_time = 0
    t_history, u_n_history, u_np_history, u_nm_history  = [cur_time], [np.array(u.sub(0).collapse().x.array)], [np.array(u.sub(1).collapse().x.array)], [np.array(u.sub(2).collapse().x.array)]
    while cur_time < t_end:
        print(f'current time: {cur_time}')

        n, converged = solver.solve(u)
        assert (converged)
        # u = problem.solve()

        cur_time += dt

        u.x.scatter_forward()
        u_last.x.array[:] = u.x.array

        x_ = np.array(msh.geometry.x[:,0])
        n_ = u.sub(0).collapse().x.array
        np_ = u.sub(1).collapse().x.array
        nm_ = u.sub(2).collapse().x.array

        t_history.append(cur_time)
        u_n_history.append(n_)
        u_np_history.append(np_)
        u_nm_history.append(nm_)

    return x_, np.array(t_history), np.array(u_n_history), np.array(u_np_history), np.array(u_nm_history)
    
    
if __name__ == '__main__':
    # Define the parameters
    x1 = 20.
    D0 = .4
    N0 = .1
    NL = .01
    km = 1.
    kp = .5
    kn = .5
    A = 7.
    sigma0 = .1
    sigmaL = .1
    n_left = N0 
    n_right = NL
    np_left = N0 * sigma0
    nm_right = NL * sigmaL
    t_end = 10.
    NTime = 200
    dt = t_end / NTime
    # number_cells = int(1/.2*x1)
    dx_ = .02#!.2
    number_cells = int(1/dx_*x1)
    relax_param = 1.
    rtol = 1e-4
    max_iter = 500
    
    # Solve the system
    x, t, n, nplus, nminus = solve_System(n_left, n_right, np_left, nm_right, D0, km, kp, kn, A, N0, NL, dt, t_end, x1=x1, relax_param=relax_param, number_cells=number_cells, max_iter=max_iter, rtol=rtol)

    print(f'n_left: {n_left}, n_right: {n_right}, np_left: {np_left}, nm_right: {nm_right}')

    # Plot the solution
    plt.figure()
    plt.plot(x, n[0,:], '--o', label='Initial')
    plt.plot(x, n[-1,:], '--o', label='Final')
    plt.legend()
    plt.grid()
    plt.xlabel('x [-]')
    plt.ylabel('$n$  [-]')
    plt.show()
    
    plt.figure()
    plt.plot(x, nplus[0,:], '--o', label='Initial')
    plt.plot(x, nplus[-1,:], '--o', label='Final')
    plt.legend()
    plt.grid()
    plt.xlabel('x [-]')
    plt.ylabel('$n_+$ [-]')
    plt.show()
    
    plt.figure()
    plt.plot(x, nminus[0,:], '--o', label='Initial')
    plt.plot(x, nminus[-1,:], '--o', label='Final')
    plt.legend()
    plt.grid()
    plt.xlabel('x [-]')
    plt.ylabel('$n_-$ [-]')
    plt.show()
