'''
Jan Habscheid
Jan.Habscheid@rwth-aachen.de
'''

import matplotlib.pyplot as plt
import numpy as np 

class System:
    '''
    System to solve the (un)coupled transient diffusion-reaction equations
    '''
    def __init__(self, L:float, D0:float, N0:float, NL:float, k_minus:float, sigma_0:float, sigma_L:float, k_p:float, k_n:float, dx:float, t_end:float, NTime:int):
        '''
        Constructor for the System class

        All parameters are in dimensionless units

        Parameters
        ----------
        L : float
            Length of the domain
        D0 : float
            Diffusion coefficient
        N0 : float
            Concentration of free particles at left boundary
        NL : float
            Concentration of free particles at right boundary
        k_minus : float
            binding rate in negative direction
        sigma_0 : float
            Degree of loading at left boundary
        sigma_L : float
            Degree of loading at right boundary
        k_p : float
            Detachment rate for particles in postive direction
        k_n : float
            Detachment rate for particles in negative direction
        dx : float
            Grid spacing
        t_end : float
            End time of simulation
        NTime : int
            Number of time steps
        '''
        # Store parameter
        self.L = L
        self.D0 = D0
        self.N0 = N0
        self.NL = NL
        self.k_minus = k_minus
        self.sigma_0 = sigma_0
        self.sigma_L = sigma_L
        self.k_p = k_p
        self.k_n = k_n
        self.dx = dx
        self.a = 1.
        self.t_end = t_end
        self.NTime = NTime
        self.dt = t_end / NTime

        self.A = None
        self.M = 1.

        # Plotting settings
        self.colors = ['blue', 'orange', 'green']
        self.ls = ['--', ':']

    def get_x(self) -> np.ndarray:
        '''
        Creates the spatial grid

        Returns
        -------
        np.ndarray
            Spatial grid
        '''
        return np.arange(self.dx * .5, self.L + .5 * self.dx, self.dx)

    def n_init_func(self, x:np.ndarray) -> np.ndarray: 
        '''
        Initial condition for the free particles

        Parameters
        ----------
        x : np.ndarray
            Spatial grid

        Returns
        -------
        np.ndarray
            Initial condition for the free particles
        '''
        n0 = self.N0 + (self.NL - self.N0) * x / self.L
        return n0
    
    def n_plus_init_func(self, x:np.array) -> np.array:
        '''
        Initial condition for the particles in the positive direction

        Parameters
        ----------
        x : np.array
            Spatial grid

        Returns
        -------
        np.array
            Initial condition for the particles in the positive direction
        '''
        n_plus = np.zeros(x.shape[0])
        n_plus[x < 0] = self.N0 * self.sigma_0
        return n_plus

    def n_minus_init_func(self, x:np.array) -> np.array:
        '''
        Initial condition for the particles in the negative direction

        Parameters
        ----------
        x : np.array
            Spatial grid

        Returns
        -------
        np.array
            Initial condition for the particles in the negative
        '''
        n_minus = np.zeros(x.shape[0])
        n_minus[x > self.L] = self.NL * self.sigma_L
        return n_minus
    
    def n_plus_ana(self, x:np.array, t:float) -> np.array:
        '''
        Analytical solution for the particles in the positive direction

        Parameters
        ----------
        x : np.array
            Spatial grid
        t : float
            Time

        Returns
        -------
        np.array
            Analytical solution for the particles in the positive direction
        '''
        return self.n_plus_init_func(x - t)

    def n_minus_ana(self, x:np.array, t:float) -> np.array:
        '''
        Analytical solution for the particles in the negative direction

        Parameters
        ----------
        x : np.array
            Spatial grid
        t : float
            Time

        Returns
        -------
        np.array
            Analytical solution for the particles in the negative direction
        '''
        return self.n_minus_init_func(x + t)

    def check_cfl(self) -> bool:
        '''
        Checks cfl-condition for advection equations

        Returns
        -------
        bool
            True if cfl-condition is satisfied, False otherwise
        '''
        return self.a * self.dt / self.dx <= 1.

    def check_Diff_stability(self) -> bool:
        '''
        Checks stability condition for diffusion equation

        Returns
        -------
        bool
            True if stability condition is satisfied, False otherwise
        '''
        return self.D0 * self.dt / self.dx**2 <= .5
    
    def check_Reaction_stability(self) -> bool:
        '''
        Checks stability condition for reaction equation

        Returns
        -------
        bool
            True if stability condition is satisfied, False otherwise
        '''
        return self.dt <= 2.0 / (1.0 + self.k_minus)

    # def check_stability(self):
    #     return self.check_cfl() and self.check_Diff_stability() and self.check_Reaction_stability()
    
    def check_stability(self) -> bool:
        '''
        Checks stability of the system

        Checks cfl-condition, diffusion stability and reaction stability
        
        Returns
        -------
        bool
            True if all stability conditions are satisfied, False otherwise
        '''
        return self.check_cfl() and self.check_Diff_stability() and self.check_Reaction_stability()
    
    def apply_bcs(self, n0:np.array, n_plus:np.array, n_minus:np.array) -> tuple:
        '''
        Applies boundary conditions (Dirichlet)

        Free particles: n(0) = N0, n(L) = NL
        Particles in positive direction: n_plus(0) = N0 * sigma_0
        Particles in negative direction: n_minus(L) = NL * sigma_L

        Parameters
        ----------
        n0 : np.array
            Free particles
        n_plus : np.array
            Particles in positive direction
        n_minus : np.array
            Particles in negative direction

        Returns
        -------
        tuple
            Free particles, Particles in positive direction, Particles in negative direction
        '''
        n0[0] = self.N0
        n0[-1] = self.NL
        n_plus[0] = self.N0 * self.sigma_0
        n_minus[-1] = self.NL * self.sigma_L
        return n0, n_plus, n_minus

    def Initialization(self) -> tuple:
        '''
        Creates initial solution

        Free particles: N0 + (NL - N0) * x / L
        Particles in positive direction: 0
        Particles in negative direction: 0

        Returns
        -------
        tuple
            Spatial grid, Free particles, Particles in positive direction, Particles in negative direction
        '''
        # Get spatial grid
        x = self.get_x()
        self.x = x

        # Initial conditions
        n_init = self.n_init_func(x)
        n_plus_init = self.n_plus_init_func(x)
        n_minus_init = self.n_minus_init_func(x)

        return x, n_init, n_plus_init, n_minus_init
    
    def update_SourceTerms(self, n:np.array, n_plus:np.array, n_minus:np.array, type:str) -> tuple:
        '''
        Applies coupling of free particles to the system

        Parameters
        ----------
        n : np.array
            Free particles
        n_plus : np.array
            Particles in positive direction
        n_minus : np.array
            Particles in negative direction
        type : str
            Type of coupling, must be one of: Standard, Exponential, Detachment

        Returns
        -------
        tuple
            Free particles, Particles in positive direction, Particles in negative direction
        '''
        n[1:-1] -= self.dt * (1 + self.k_minus) * n[1:-1]
        n_plus[1:] += self.dt * n[1:]
        n_minus[:-1] += self.dt * self.k_minus * n[:-1]
        return n, n_plus, n_minus
    
    def update_Reaction(self, n:np.array, n_plus:np.array, n_minus:np.array, type:str) -> tuple:
        '''
        Applies coupling of particles moving along tractions to the system

        Parameters
        ----------
        n : np.array
            Free particles
        n_plus : np.array
            Particles in positive direction
        n_minus : np.array
            Particles in negative direction
        type : str
            Type of coupling, must be one of: Standard, Exponential, Detachment

        Returns
        -------
        tuple
            Free particles, Particles in positive direction, Particles in negative direction
        '''
        if type in ['Standard', 'Exponential']:
            k_p_n, k_p_np = self.k_p, self.k_p
            k_n_n, k_n_nm = self.k_n, self.k_n
        elif type == 'Detachment':
            k_p_n = self.k_p * np.exp(self.B * n_plus[1:-1])
            k_n_n = self.k_n * np.exp(self.B * n_minus[1:-1])
            k_p_np = self.k_p * np.exp(self.B * n_plus[1:])
            k_n_nm = self.k_n * np.exp(self.B * n_minus[:-1])
        n[1:-1] += self.dt * (
            k_p_n * n_plus[1:-1] \
            + k_n_n * n_minus[1:-1] \
        )
        n_plus[1:] += self.dt * (
            - k_p_np * n_plus[1:]
        )
        n_minus[:-1] += self.dt * (
            - k_n_nm * n_minus[:-1]
        )
        return n, n_plus, n_minus
    
    def set_A(self, A:float) -> None:
        '''
        Sets the parameter A for the exponential function

        Parameters
        ----------
        A : float
            Parameter for the exponential function
        '''
        self.A = A
    
    def f(self, u:np.array, sign:str) -> np.array:
        '''
        Exponential closure for the velocity

        u+ = u * exp(-A * u)
        u- = -u * exp(-A * u)

        Parameters
        ----------
        u : np.array
            number densities
        sign : str
            Sign of the velocity, must be one of: "+", "-"

        Returns
        -------
        np.array
            Velocity

        Raises
        ------
        ValueError
            If sign is not one of: "+", "-"
        '''
        match sign:
            case '+': return u * np.exp(-self.A * u)
            case '-': return -u * np.exp(-self.A * u)
        raise ValueError('Sign must be one of: "+", "-"')
    
    def Upwind_n_plus(self, n_plus_old:np.array, type:str='Standard') -> tuple:
        '''
        Upwind discretization for the positive direction

        Parameters
        ----------
        n_plus_old : np.array
            Particles in positive direction
        type : str, optional
            Type of discretization, by default 'Standard'

        Returns
        -------
        tuple
            Upwind discretization for the positive direction
        '''
        # Normal upwind discretization
        if type in ['Standard', 'Detachment']:
            return n_plus_old[1:], n_plus_old[:-1]
        # specific discretization for nonlinear exponential function
        elif type=='Exponential':
            F = np.zeros_like(n_plus_old)
            F[:-1] = .5 * (
                self.f(n_plus_old[:-1], sign='+') + self.f(n_plus_old[1:], sign='+')
                - self.M * (n_plus_old[1:] - n_plus_old[:-1])
            )
            F[-1] = self.f(n_plus_old[-1], sign='+')  # Ensure boundary condition correctness
            return F[1:], F[:-1]

    def Upwind_n_minus(self, n_minus_old:np.array, type:str='Standard') -> tuple:
        '''
        Upwind discretization for the negative direction

        Parameters
        ----------
        n_minus_old : np.array
            Particles in negative direction
        type : str, optional
            Type of discretization, by default 'Standard'

        Returns
        -------
        tuple
            Upwind discretization for the negative direction
        '''
        # Normal upwind discretization
        if type in ['Standard', 'Detachment']:
            return n_minus_old[:-1], n_minus_old[1:]
        # specific discretizatin for nonlinera exponential function
        elif type=='Exponential':
            F = np.zeros_like(n_minus_old)
            F[1:] = .5 * (
                self.f(n_minus_old[:-1], sign='-') + self.f(n_minus_old[1:], sign='-')
                - self.M * (n_minus_old[1:] - n_minus_old[:-1])
            )
            F[0] = self.f(n_minus_old[0], sign='-')  # Ensure boundary condition correctness
            return F[1:], F[:-1]
        
    def Upwind(self, n_plus_old:np.array, n_minus_old:np.array, type:str='Standard') -> tuple:
        '''
        Upwind Discretization for the system

        Combines Upwind discretization for the positive and negative direction

        Parameters
        ----------
        n_plus_old : np.array
            Particles in positive direction
        n_minus_old : np.array
            Particles in negative direction
        type : str, optional
            Type of discretization, by default 'Standard'

        Returns
        -------
        tuple
            Upwind discretization for the positive and negative direction
        '''
        return self.Upwind_n_plus(n_plus_old, type), self.Upwind_n_minus(n_minus_old, type)
    
    def solve_System(self, exchange='None', check_propagation=False, check_stability=True, atol=1e-3, rtol=1e-3, type='Standard') -> None:
        '''
        Solves the system of equations

        Time is discretized with a forward Euler scheme
        Diffusion is discretized with a central difference scheme
        Advection is discretized with an upwind scheme

        First, the system is initialized
        Then, the system is solved for each time step
        For each time step, first the derivatives are applied and after, the source terms are added, based on those "intermediate" solutions

        Parameters
        ----------
        exchange : str, optional
            Type of exchange, must be one of: None, SourceTerms, ReactionTerms, FullTerms, by default 'None'
        check_propagation : bool, optional
            If True, stops when System reached stationary case, by default False
        check_stability : bool, optional
            If True, checks stability of the system, by default True
        atol : _type_, optional
            Absolute Tolerance for checking stationary behavior, by default 1e-3
        rtol : _type_, optional
            Relative Tolerance for checking stationary behavior, by default 1e-3
        type : str, optional
            Type of discretization, must be one of: Standard, Exponential, Detachment, by default 'Standard

        Raises
        ------
        ValueError
            Stability check throws an error
        ValueError
            Exchange method not recognized and must be one of: None, SourceTerms, ReactionTerms, FullTerms
        '''
        # Initialization
        x, n_init, n_plus_init, n_minus_init = self.Initialization()
        n = n_init.copy()
        n_plus = n_plus_init.copy()
        n_minus = n_minus_init.copy()

        # Solution vector
        n_vec = [n.copy()]
        n_plus_vec = [n_plus_init.copy()]
        n_minus_vec = [n_minus_init.copy()]

        # Flux
        nx = (n[2] - 2 * n[1] + n[0]) / 2 * self.dx # Central difference
        J_flux = [n_plus.copy() + n_minus.copy() - self.D0 * nx]

        t = .0
        t_vec = [t]
        # Check stability
        stability = self.check_stability()
        if check_stability and not stability:
            raise ValueError('Stability conditions not satisfied')

        # Time loop 
        for j in range(self.NTime):
            # Check for propagation
            if check_propagation:
                n_m_left = n_minus[0]
                n_p_right = n_plus[-1]
                if np.isclose(n_p_right, self.N0 * self.sigma_0, rtol=rtol, atol=atol) and np.isclose(n_m_left, self.NL * self.sigma_L, rtol=rtol, atol=atol):
                    print(f'Convergence reached after {iter} iterations')
                    print(f'Needed time for full propagation: {t}')
                    print(f'Tolerance is: atol={atol}, rtol={rtol}')
                    self.t_propagation = t
                    break

            # Store old values
            n_old, n_plus_old, n_minus_old = n_vec[-1], n_plus_vec[-1], n_minus_vec[-1]
            # Get Upwind
            (upwind_n_plus_1, upwind_n_plus_2), (upwind_n_minus_1, upwind_n_minus_2) = self.Upwind(n_plus_old, n_minus_old, type=type)
            # no exchange
            n[1:-1] = n_old[1:-1] + self.dt * (
                self.D0 / self.dx**2 * (
                    n_old[2:] - 2 * n_old[1:-1] + n_old[:-2]
                )
            )
            n_plus[1:] = n_plus_old[1:] - self.dt * (
                1 / self.dx  * (upwind_n_plus_1 - upwind_n_plus_2)
            )
            n_minus[:-1] = n_minus_old[:-1] - self.dt * (
                1 /self. dx  * (upwind_n_minus_1 - upwind_n_minus_2)
            )

            match exchange:
                case 'None':
                    pass
                case 'SourceTerms':
                    n, n_plus, n_minus = self.update_SourceTerms(n, n_plus, n_minus, type)
                case 'ReactionTerms':
                    n, n_plus, n_minus = self.update_Reaction(n, n_plus, n_minus, type)
                case 'FullTerms':
                    # n, n_plus, n_minus = self.update_Reaction(n, n_plus, n_minus, type)
                    n, n_plus, n_minus = self.update_SourceTerms(n, n_plus, n_minus, type)
                    n, n_plus, n_minus = self.update_Reaction(n, n_plus, n_minus, type)
                case _:
                    raise ValueError('Exchange method not recognized, must be one of: None, SourceTerms, ReactionTerms, FullTerms')

            # Apply BCs
            n, n_plus, n_minus = self.apply_bcs(n, n_plus, n_minus)

            # Store
            n_vec.append(n.copy())
            n_plus_vec.append(n_plus.copy())
            n_minus_vec.append(n_minus.copy())
            J_flux.append(n_plus.copy() + n_minus.copy() - self.D0 * nx)

            t_vec.append(t)
            t += self.dt

        # Append current solution to solution vector
        t_vec = np.array(t_vec)
        n_vec = np.array(n_vec)
        n_plus_vec = np.array(n_plus_vec)
        n_minus_vec = np.array(n_minus_vec)
        J_flux = np.array(J_flux)

        # Store solution
        self.t_vec = t_vec
        self.n_vec = n_vec
        self.n_plus_vec = n_plus_vec
        self.n_minus_vec = n_minus_vec
        self.J_flux = J_flux

    def plot_initial_conditions(self) -> tuple:
        '''
        Plots initial solution

        3x1 plot for free particles, particles in positive direction and particles in negative direction

        Returns
        -------
        tuple
            Figure and Axes
        '''
        x, n_init, n_plus_init, n_minus_init = self.Initialization()
        fig, axs = plt.subplots()
        fig.suptitle('Initial Solution')
        axs.plot(self.x, n_init, label='$n$')
        axs.plot(x, n_plus_init, label='$n_+$')
        axs.plot(x, n_minus_init, label='$n_-$')
        axs.grid()
        axs.legend()
        fig.tight_layout()
        return fig, axs
    
    def plot_analytical_solution(self, t:float) -> tuple:
        '''
        Plots analytical solution

        2x1 plot for particles in positive and negative direction
        Diffusion equation already satisfies the analytical solution

        Parameters
        ----------
        t : float
            Time

        Returns
        -------
        tuple
            Figure and Axes
        '''
        x, _, _, _ = self.Initialization()
        fig, axs = plt.subplots()
        fig.suptitle(f'Analytical Solution at t={t}')
        axs.plot(x, self.n_plus_ana(x, t), '--o', label='$n_+$')
        axs.plot(x, self.n_minus_ana(x, t), '--o', label='$n_-$')
        axs.grid()
        fig.legend()
        fig.tight_layout()
        return fig, axs
    
    def plot_solution(self, timeindex:int, analytical=False, initial=False) -> tuple:
        '''
        Plots solution in 1-D plot at timeindex

        3x1 plot for free particles, particles in positive direction and particles in negative direction

        Parameters
        ----------
        timeindex : int
            Time index to plot
        analytical : bool, optional
            If True, adds analytical solution to plot, by default False
        initial : bool, optional
            If True, adds initial solution to plot, by default

        Returns
        -------
        tuple
            Figure and Axes
        '''
        # 1-D plot
        fig, axs = plt.subplots(ncols=3, figsize=(15, 5))
        fig.suptitle(f'Solution at t={self.t_end}')

        # Plot number densities
        axs[0].plot(self.x, self.n_vec[-1], color=self.colors[0])
        axs[1].plot(self.x, self.n_plus_vec[-1], color=self.colors[0], label='Numerical')
        axs[2].plot(self.x, self.n_minus_vec[-1], color=self.colors[0])
        [ax.grid() for ax in axs]
        
        # Labeling
        [ax.set_title(t) for ax, t in zip(axs, ['$n$', '$n_+$', '$n_-$'])]
        [ax.set_xlabel('$x$ [-]') for ax in axs]
        [ax.set_ylabel(t) for ax, t in zip(axs, ['$n [-]', '$n_+$ [-]', '$n_-$ [-]'])]
        
        # Add analytical solution
        if analytical:
            axs[1].plot(self.x, self.n_plus_ana(self.x, self.t_vec[timeindex]), '--', color=self.colors[0], label='Analytical')
            axs[2].plot(self.x, self.n_minus_ana(self.x, self.t_vec[timeindex]), '--', color=self.colors[0])
        # Add initial solution
        if initial:
            axs[0].plot(self.x, self.n_init_func(self.x), '--', color=self.colors[0], label='Initial')
            axs[1].plot(self.x, self.n_plus_init_func(self.x), ':', color=self.colors[0])
            axs[2].plot(self.x, self.n_minus_init_func(self.x), ':', color=self.colors[0])
        # Add legend
        if analytical or initial:
            fig.legend()
        fig.tight_layout()

        return fig, axs
    
    def plot_solution_contourf(self) -> tuple:
        '''
        Plots solution in 2-D contourf plot

        3x1 plot for free particles, particles in positive direction and particles in negative direction

        Returns
        -------
        tuple
            Figure and Axes
        '''
        # contourfigures over time for each
        fig, axs = plt.subplots(ncols=3, figsize=(15, 5))
        fig.suptitle(f'Solution for $t \in [0,{self.t_end}]$')

        # contourf
        cfig0 = axs[0].contourf(self.x, self.t_vec, self.n_vec, cmap='coolwarm')
        cfig1 = axs[1].contourf(self.x, self.t_vec, self.n_plus_vec, cmap='coolwarm')
        cfig2 = axs[2].contourf(self.x, self.t_vec, self.n_minus_vec, cmap='coolwarm')

        # colorbar
        fig.colorbar(cfig0, ax=axs[0])
        fig.colorbar(cfig1, ax=axs[1])
        fig.colorbar(cfig2, ax=axs[2])
        
        # Labeling
        [ax.grid() for ax in axs]
        [ax.set_title(t) for ax, t in zip(axs, ['$n$', '$n_+$', '$n_-$'])]
        [ax.set_xlabel('$x$ [-]') for ax in axs]
        [ax.set_ylabel('$t$ [-]') for ax in axs]
        fig.tight_layout()

        return fig, axs
    
    def plot_solution_3d(self) -> tuple:
        '''
        Plots solution in 3-D plot

        3x1 plot for free particles, particles in positive direction and particles in negative direction

        Returns
        -------
        tuple
            Figure and Axes
        '''
        # 3-D plot
        fig = plt.figure(figsize=(15, 5))
        axs = [fig.add_subplot(131, projection='3d'), fig.add_subplot(132, projection='3d'), fig.add_subplot(133, projection='3d')]
        fig.suptitle(f'Solution for $t \in [0,{self.t_end}]$')

        # Plot
        X, T = np.meshgrid(self.x, self.t_vec)
        axs[0].plot_surface(X, T, self.n_vec, cmap='coolwarm')
        axs[1].plot_surface(X, T, self.n_plus_vec, cmap='coolwarm')
        axs[2].plot_surface(X, T, self.n_minus_vec, cmap='coolwarm')

        # Labeling
        [ax.set_title(t) for ax, t in zip(axs, ['$n$', '$n_+$', '$n_-$'])]
        [ax.set_xlabel('$x$ [-]') for ax in axs]
        [ax.set_ylabel('$t$ [-]') for ax in axs]
        [ax.set_zlabel(t) for ax, t in zip(axs, ['$n$ [-]', '$n_+$ [-]', '$n_-$ [-]'])]

        fig.tight_layout()
        return fig, axs
    
    def plot_flux(self) -> tuple:
        '''
        Plots flux

        1x1 contourf for flux

        Returns
        -------
        tuple
            Figure and Axes
        '''
        # 1-D plot
        fig, axs = plt.subplots()
        fig.suptitle(f'Flux for $t \in [0,{self.t_end}]$')

        # contourf with colorbar
        cfig = axs.contourf(self.x, self.t_vec, self.J_flux, cmap='coolwarm')
        fig.colorbar(cfig, ax=axs)

        # Labeling
        axs.grid()
        axs.set_xlabel('x [-]')
        axs.set_ylabel('t [-]')
        fig.tight_layout()

        return fig, axs
